import xmltodict
import json
import re

from tokens_dict import TokensDict

from gensim.models.phrases import Phrases, ENGLISH_CONNECTOR_WORDS, Phraser

class CapecSpecsDict(TokensDict):
    def __init__(self, capec_file, training_corpus_file):
        super().__init__()
        
        self.__capec_names = []
        self.__capec_descriptions = []
        self.__capec_execution_flows = []
        self.__capec_mitigations = []
        self.__capec_tokens = []
        self.__capec_tokens_lemm = []
        self.__system_spec_tokens = []
        self.__system_spec_tokens_lemm = []
        self.__training_tokens = []
        # self.__bigram_model = None
        # self.__trigram_model = None

        self.parse_capecs(capec_file)
        self.parse_training_corpus(training_corpus_file)

    def get_capec_names(self):
        return self.__capec_names

    def get_capec_descs(self):
        return self.__capec_descriptions
    
    def get_capec_execution_flows(self):
        return self.__capec_execution_flows
    
    def get_capec_mitigations(self):
        return self.__capec_mitigations
    
    def get_capec_tokens(self):
        return self.__capec_tokens
    
    def get_capec_tokens_lemm(self):
        return self.__capec_tokens_lemm
    
    def get_system_spec_tokens(self):
        return self.__system_spec_tokens
    
    def get_system_spec_tokens_lemm(self):
        return self.__system_spec_tokens_lemm
    
    def get_training_tokens(self):
        return self.__training_tokens
    
    def get_training_word_doc_count(self):
        word_document_count = {}  # Dictionary to store word-document counts

        # Iterate over each document in the corpus
        for document in self.__training_tokens:
            unique_words_in_doc = set(document)

            # Update the word-document count for each unique word in the document
            for word in unique_words_in_doc:
                # Increment the count for the word if it exists, otherwise initialize it to 1
                word_document_count[word] = word_document_count.get(word, 0) + 1

        sorted_counts = sorted(word_document_count.items(), key=lambda x: (-x[1], x[0]))

        return sorted_counts
    
    def parse_capecs(self, capec_file):
        capec_descriptions = []
        capec_names = []
        capec_execution_flows = []
        capec_mitigations = []
        
        with open(capec_file, 'r') as xml_file:
            xml_data = xml_file.read()
            xml_data = re.sub(r"<xhtml:(.*?)>", "", xml_data)
            xml_data = re.sub(r"</xhtml:(.*?)>", "", xml_data)
            capec_dict = xmltodict.parse(xml_data)

        attack_patterns = capec_dict['Attack_Pattern_Catalog']['Attack_Patterns']['Attack_Pattern']

        for attack_pattern in attack_patterns:
            if (attack_pattern['@Abstraction'] == 'Standard' or \
                attack_pattern['@Abstraction'] == 'Detailed') and \
                attack_pattern['@Status'] != 'Obsolete' and \
                attack_pattern['@Status'] !='Deprecated' and \
                "Execution_Flow" in attack_pattern:
                
                capec_names.append(attack_pattern['@Name'])
                
                description = ""
                description = attack_pattern["Description"]

                if "Extended_Description" in attack_pattern:
                    description = description + " " + attack_pattern["Extended_Description"]
                
                capec_descriptions.append(description)
                
                attack_steps_list = []
                attack_steps = attack_pattern["Execution_Flow"]["Attack_Step"]
                
                if isinstance(attack_steps, list):
                    for idx, attack_step in enumerate(attack_steps):
                        cleaned_attack_step = "Step " + str(idx + 1) + ". " + \
                            super().clean_text(attack_step["Description"])
                        attack_steps_list.append(cleaned_attack_step)
                else:
                    cleaned_attack_step = "Step 1. " + super().clean_text(attack_steps["Description"])
                    attack_steps_list.append(cleaned_attack_step)

                capec_execution_flows.append(attack_steps_list)

                mitigations_list = []

                if "Mitigations" in attack_pattern:
                    mitigations = attack_pattern["Mitigations"]["Mitigation"]

                    if isinstance(mitigations, list):
                        for mitigation in mitigations:
                            cleaned_mitigation = super().clean_text(mitigation)
                            mitigations_list.append(cleaned_mitigation)
                    else:
                        cleaned_mitigation = super().clean_text(mitigations)
                        mitigations_list.append(cleaned_mitigation)

                capec_mitigations.append(mitigations_list)

        capec_descriptions = super().clean_texts(capec_descriptions)

        self.__capec_names = capec_names
        self.__capec_descriptions = capec_descriptions
        self.__capec_execution_flows = capec_execution_flows
        self.__capec_mitigations = capec_mitigations

        capec_tokens = super().tokenize_text(capec_descriptions)
        capec_tokens = super().lowercase_tokens(capec_tokens)
        capec_tokens = super().remove_numbers(capec_tokens)
        
        capec_tokens_lemm = super().lemmatize_tokens(capec_tokens)

        self.__capec_tokens = capec_tokens
        self.__capec_tokens_lemm = capec_tokens_lemm

        # bigram = Phrases(training_tokens, min_count=7, threshold=18, 
        #                  connector_words=ENGLISH_CONNECTOR_WORDS)
        # self.__bigram_model = Phraser(bigram)

        # trigram = Phrases(bigram[training_tokens], min_count=16, threshold=26, 
        #                   connector_words=ENGLISH_CONNECTOR_WORDS)
        # self.__trigram_model = Phraser(trigram)

        # # Apply the bigram model to each document in the training tokens corpus
        # training_tokens = [self.__bigram_model[doc] for doc in training_tokens]
        # # Apply the trigram model to the resulting bigram training tokens corpus
        # training_tokens = [self.__trigram_model[doc] for doc in training_tokens]

    def parse_training_corpus(self, training_corpus_file):
        unduped_training_tokens = []
        
        # Open the JSON file
        with open(training_corpus_file, 'r') as file:
            # Load the JSON data
            data = json.load(file)

        for source in ["MS-Bulletin", "Metasploit", "NVD"]:
            for _, value in data[source].items():
                corpora_tokens = []

                for word_label in value:
                    word = word_label[0]

                    if word.isalnum():
                        corpora_tokens.append(word)

                unduped_training_tokens.append(corpora_tokens)

        unduped_training_tokens = super().lowercase_tokens(unduped_training_tokens)
        unduped_training_tokens = super().remove_numbers(unduped_training_tokens)

        unduped_training_tokens += self.__capec_tokens

        training_tokens_dict = {}

        # Iterate through each array in the outer array
        for index, token_set in enumerate(unduped_training_tokens):
            # Sort the elements of the array to ignore the order
            sorted_token_set = sorted(token_set)

            # Convert the sorted array to a tuple to make it hashable
            token_set_tuple = tuple(sorted_token_set)

            # Compute the hash value of the sorted array
            token_set_hash = hash(token_set_tuple)

            # Check if the hash value already exists in the dictionary
            if token_set_hash not in training_tokens_dict:
                # If the hash value doesn't exist, add the sorted array to the dictionary
                training_tokens_dict[token_set_hash] = index            

        training_tokens = [unduped_training_tokens[index] for index in list(training_tokens_dict.values())]

        self.__training_tokens = training_tokens

    def parse_system_specs(self, spec_file):
        with open(spec_file, 'r') as file:
            # Read the contents of the file
            file_contents = file.read()

        system_spec_texts = super().clean_texts([file_contents])

        system_spec_tokens = super().tokenize_text(system_spec_texts)
        system_spec_tokens = super().lowercase_tokens(system_spec_tokens)
        system_spec_tokens = super().remove_numbers(system_spec_tokens)
        
        system_spec_tokens_lemm = super().lemmatize_tokens(system_spec_tokens)

        self.__system_spec_tokens = system_spec_tokens[0]
        self.__system_spec_tokens_lemm = system_spec_tokens_lemm[0]