import requests
import os

from capec_specs_dict import CapecSpecsDict

from gensim.corpora import Dictionary
from gensim.models import LdaModel
from gensim.matutils import hellinger
from gensim import similarities
from gensim.models import TfidfModel
from sklearn.utils import shuffle

def get_capec_file(abs_path):
    # Send a GET request to the URL
    response = requests.get("https://capec.mitre.org/data/xml/capec_latest.xml")
    
    # Check if the request was successful (status code 200)
    if response.status_code == 200:
        # Open the file in binary write mode and write the content
        with open(abs_path + "capec_latest.xml", 'wb') as file:
            file.write(response.content)

    return response

def trace_capecs(abs_path):
    try:
        tokens_dicti = CapecSpecsDict(abs_path + "capec_latest.xml", abs_path + 'full_corpus.json')
            
        capec_names = tokens_dicti.get_capec_names()
        capec_descs = tokens_dicti.get_capec_descs()
        capec_attack_steps = tokens_dicti.get_capec_execution_flows()
        capec_mitigations = tokens_dicti.get_capec_mitigations()
        capec_tokens = tokens_dicti.get_capec_tokens()
        capec_tokens_lemm = tokens_dicti.get_capec_tokens_lemm()

        tokens_dicti.parse_system_specs(abs_path + "system_specs.txt")
        spec_tokens = tokens_dicti.get_system_spec_tokens()
        spec_tokens_lemm = tokens_dicti.get_system_spec_tokens_lemm()

        # Create a dictionary from the corpus documents
        capec_lemm_dictionary = Dictionary(capec_tokens_lemm)

        # Create a bag-of-words corpus from the corpus documents
        capec_lemm_corpus = [capec_lemm_dictionary.doc2bow(doc) for doc in capec_tokens_lemm]

        # Create a TF-IDF model
        tfidf_model = TfidfModel(capec_lemm_corpus, id2word=capec_lemm_dictionary)

        # Convert the checked document to TF-IDF representation
        spec_lemm_bow = capec_lemm_dictionary.doc2bow(spec_tokens_lemm)
        spec_tfidf = tfidf_model[spec_lemm_bow]

        # Create a similarity index
        index = similarities.MatrixSimilarity(tfidf_model[capec_lemm_corpus], num_features=len(capec_lemm_dictionary))

        # Calculate similarity scores
        similarity_scores = index[spec_tfidf]

        # Combine similarity scores with document indices
        doc_similarity_pairs = list(enumerate(similarity_scores))

        # Sort documents by similarity score
        sorted_doc_similarity_pairs = sorted(doc_similarity_pairs, key=lambda x: x[1], reverse=True)

        # Print the top N most similar documents
        # for idx, score in sorted_doc_similarity_pairs:
        #     print(score)
        #     print(capec_descs[idx])

        if sorted_doc_similarity_pairs[1][1] > 0:
            training_tokens = tokens_dicti.get_training_tokens()
            training_tokens = shuffle(training_tokens)
            
            training_dictionary = Dictionary(training_tokens)
            training_dictionary.filter_extremes(no_below=2, no_above=0.3)

            training_corpus = [training_dictionary.doc2bow(text) for text in training_tokens]

            lda = LdaModel(corpus=training_corpus,
                            num_topics=18,
                            id2word=training_dictionary,
                            passes=12,
                            alpha="auto",
                            eta="auto",
                            iterations=6,   
                            update_every=1,
                            decay=0.745786579640169,
                            offset=3.20092862358715
                            )
            
            system_specs_bow = training_dictionary.doc2bow(spec_tokens)
            # system_spec_words = set(training_dictionary[word] for word, _ in system_specs_bow) 

            # Extract the top words for each topic from the new document
            # system_spec_topics = lda.get_document_topics(system_specs_bow)
            # system_spec_topic_words = []

            # for topic_id, prob in system_spec_topics:
            #     top_words = lda.show_topic(topic_id, topn=len(system_spec_words))
            #     top_words = [word for word, _ in top_words if word in system_spec_words]
            #     system_spec_topic_words.append(top_words)

            #  # Print the top words for each topic
            # for i, top_words in enumerate(system_spec_topic_words):
            #     print(f"Topic {i}: {', '.join(top_words)}")

            system_spec_topic_dist = lda[system_specs_bow]

            difference_scores = []

            for index, capec_token in enumerate(capec_tokens):
                capec_desc_bow = training_dictionary.doc2bow(capec_token)
                capec_spec_topic_dist = lda[capec_desc_bow]

                score = [index, hellinger(capec_spec_topic_dist, system_spec_topic_dist)]

                difference_scores.append(score)

            # Sort the arrays based on the value of the second index in each array
            difference_scores = sorted(difference_scores, key=lambda x: x[1], reverse=False)

            with open(abs_path + 'traced_capecs.txt', 'w') as output:
                for score in difference_scores:
                    output.write(f'Confidence score: {100 - int(100 * score[1])}%\n')
                    output.write(f'Name: {capec_names[score[0]]}\n')
                    output.write(f'Description:\n')
                    output.write(f'{capec_descs[score[0]]}\n')
                    output.write(f'Attack Steps:\n')
                    for attack_step in capec_attack_steps[score[0]]:
                        output.write(f'{attack_step}\n')
                    if capec_mitigations[score[0]]:
                        output.write(f"Mitigations:\n")
                        for mitigation in capec_mitigations[score[0]]:
                            output.write(f'{mitigation}\n')
                    output.write('\n')
        else:
            with open(abs_path + 'traced_capecs.txt', 'w') as output:
                output.write("No attack patterns were able to be identified with the provided system specifications. " + 
                    "Please provide the system specifications with additional details.\n")
    except Exception as e:
        with open(abs_path + 'traced_capecs.txt', 'w') as output:
                output.write(e)

if __name__ == "__main__": 
    abs_path = os.path.abspath(__file__)
    last_slash_index = abs_path.rfind("/")
    abs_path = abs_path[:last_slash_index + 1]
    
    response = get_capec_file(abs_path)

    if response.status_code == 200:
        trace_capecs(abs_path)
    else:
        with open(abs_path + 'traced_capecs.txt', 'w') as output:
                output.write(f"Failed to download the list of CAPECs from MITRE. Status code: {response.status_code}.")
