import tqdm
import logging
import numpy as np
import os.path
import requests

from capec_specs_dict import CapecSpecsDict

from gensim.corpora import Dictionary
from gensim.models import LdaModel
from gensim.models.callbacks import CoherenceMetric, PerplexityMetric, ConvergenceMetric
from bayes_opt import BayesianOptimization
from pandas import DataFrame
from bayes_opt.logger import JSONLogger
from bayes_opt.event import Events
from sklearn.model_selection import KFold
from functools import partial
    
# Train model 
def train_lda(training_tokens, dictionary, num_topics, passes, iterations, update_every, 
              training_logs, training_results_csv, eval_every, decay, offset):
    model_dict = {'Topics': [],
                'Passes': [],
                'Iterations': [],
                'Update_Every': [],
                'Decays': [],
                'Offsets': [],
                'Avg_Coherences': [],
                'Avg_Perplexities': [],
                'Avg_Convergences': [],
                'Dev_Coherences': [],
                'Dev_Perplexities': [],
                'Dev_Convergences': [],
                'Epochs': []}
    
    # Initialize k-fold cross-validation
    k = 4
    kf = KFold(n_splits=k, shuffle=True, random_state=1)

    # Initialize a list to store evaluation metrics for each fold
    coherence_scores = []
    perplexity_scores = []
    convergence_scores = []

    k_index = 1

    print("Training with", num_topics, "topics,", passes, 
          "max passes,", iterations, "max iterations,", 
          update_every, "update every,", decay, "decay,", offset, "offset:", sep=" ")

    pbar = tqdm.trange(k)

    # Iterate over each fold
    for train_index, test_index in kf.split(training_tokens):
        train_tokens = [training_tokens[i] for i in train_index]
        val_tokens = [training_tokens[i] for i in test_index]
        
        train_corpus = [dictionary.doc2bow(doc) for doc in train_tokens]
        val_corpus = [dictionary.doc2bow(doc) for doc in val_tokens]

        # Configure logging
        logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=logging.DEBUG,
                        filename=training_logs)
        
        logging.info(
            "Training with %s topics, %s max passes, %s max iterations, %s update every, %s decay, %s offset at %s fold:", 
            num_topics, passes, iterations, update_every, decay, offset, k_index)
        
        coherence_metric = CoherenceMetric(
            texts=train_tokens, coherence='c_npmi', window_size=10)
        perplexity_metric = PerplexityMetric(corpus=val_corpus)
        convergence_metric = ConvergenceMetric(distance="hellinger")
        
        lda_model = LdaModel(corpus=train_corpus,
                            num_topics=num_topics,
                            id2word=dictionary,
                            passes=passes,
                            alpha="auto",
                            eta="auto",
                            random_state=1,
                            iterations=iterations,   
                            update_every=update_every,
                            decay=decay,
                            offset=offset,
                            eval_every=eval_every,
                            callbacks=[perplexity_metric, coherence_metric, convergence_metric],
                            chunksize=2000)
        
        logging.shutdown()

        coherence_scores.append(lda_model.metrics['Coherence'])
        perplexity_scores.append(lda_model.metrics['Perplexity'])
        convergence_scores.append(lda_model.metrics['Convergence'])

        k_index += 1

        pbar.update(1)
    
    pbar.close()

    for epoch, _ in enumerate(coherence_scores[0]):
        model_dict['Topics'].append(num_topics)
        model_dict['Passes'].append(passes)
        model_dict['Iterations'].append(iterations)
        model_dict['Update_Every'].append(update_every)
        model_dict['Decays'].append(decay)
        model_dict['Offsets'].append(offset)
        model_dict['Epochs'].append(epoch + 1)

    coherence_score_averages = np.mean(coherence_scores, axis=0)
    perplexity_score_averages = np.mean(perplexity_scores, axis=0)
    convergence_score_averages = np.mean(convergence_scores, axis=0)

    # Compute the absolute deviation of each element from its corresponding column mean
    coherence_absolute_deviations = np.abs(coherence_scores - coherence_score_averages)
    perplexity_absolute_deviations = np.abs(perplexity_scores - perplexity_score_averages)
    convergence_absolute_deviations = np.abs(convergence_scores - convergence_score_averages)

    # Compute the average of these absolute deviations for each column
    coherece_absolute_average_deviations = np.mean(coherence_absolute_deviations, axis=0)
    perplexity_absolute_average_deviations = np.mean(perplexity_absolute_deviations, axis=0)
    convergence_absolute_average_deviations = np.mean(convergence_absolute_deviations, axis=0)

    model_dict['Avg_Coherences'] = coherence_score_averages
    model_dict['Avg_Perplexities'] = perplexity_score_averages
    model_dict['Avg_Convergences'] = convergence_score_averages
    model_dict['Dev_Coherences'] = coherece_absolute_average_deviations
    model_dict['Dev_Perplexities'] = perplexity_absolute_average_deviations
    model_dict['Dev_Convergences'] = convergence_absolute_average_deviations

    model_df = DataFrame.from_dict(model_dict)

    if os.path.isfile(training_results_csv):
        model_df.to_csv(training_results_csv, mode='a', header=False)
    else:
        model_df.to_csv(training_results_csv, mode='w', header=True)

    return coherence_score_averages[len(coherence_score_averages) - 1]
    # return -perplexity_score_averages[len(perplexity_score_averages) - 1]

def train_lda_wrapper(training_tokens, dictionary, num_topics, passes, iterations, update_every, 
              training_logs, training_results_csv, eval_every, decay=0.5, offset=1):
    num_topics = int(num_topics)
    passes = int(passes)
    iterations = int(iterations)
    update_every = int(update_every)

    return train_lda(training_tokens, dictionary, num_topics, passes, iterations, update_every, 
              training_logs, training_results_csv, eval_every, decay, offset)

def get_capec_file():
    # Send a GET request to the URL
    response = requests.get("https://capec.mitre.org/data/xml/capec_latest.xml")
    
    # Check if the request was successful (status code 200)
    if response.status_code == 200:
        # Open the file in binary write mode and write the content
        with open("capec_latest.xml", 'wb') as file:
            file.write(response.content)

    return response

if __name__ == "__main__": 
    response = get_capec_file()

    if response.status_code == 200:
        tokens_dicti = CapecSpecsDict("capec_latest.xml", 'full_corpus.json')

        training_tokens = tokens_dicti.get_training_tokens()

        # Create dictionary and train/test corpus
        dictionary = Dictionary(training_tokens)
        dictionary.filter_extremes(no_below=2, no_above=0.3)

        train_lda_partial = partial(train_lda_wrapper, 
                                    training_tokens=training_tokens,
                                    dictionary=dictionary,
                                    training_logs="lda_training4.log",
                                    training_results_csv="lda_training_results4.csv", 
                                    eval_every=10
                                    )

        # train_lda_partial(num_topics=181, 
        #                   passes=20, iterations=len(tokens_dicti.get_capec_names()), update_every=1, decay=0.7, offset=10)

        optimizer = BayesianOptimization(
            f=train_lda_partial,
            pbounds={'num_topics': (2, 30), 
                    'passes': (10, 20),
                    'iterations': (1, 100),
                    'update_every': (1, 1),
                    'decay': (0.5, 1),
                    'offset': (1, 16)},
            verbose=2,
            random_state=1,
        )

        optimizer.set_gp_params(alpha=1e-3)

        logger = JSONLogger(path="./lda_optimization4_log")
        optimizer.subscribe(Events.OPTIMIZATION_STEP, logger)

        optimizer.maximize(init_points=10, n_iter=50)
        print(optimizer.max)
    else:
        print(f"Failed to download the list of CAPECs from MITRE. Status code: {response.status_code}.")